/* minimization of M(natural)-convex function
 *  This is an implementation of steepest descent algorithm.
 *  The calculation cost is O(dimension^2) per step.
 *  cf. K. Murota: Discrete convex analysis, 2003, pp.281--283.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "odicon.h"

/* returns FALSE if no improvment was found. */
/* returns TRUE if improvment was found.
 '*plus' and '*minus' are the indexes of 'x[]'. */
static int best_improve(int dim, double func(int, int*), int x[],
                        int *plus, int *minus, int natural) {
    int i, j, improved = FALSE;
    double f, min_f;

    iteration++;
    min_f = func(dim, x);
    for (i=0; i<dim+natural; i++) {
        x[i]--;
        for (j=0; j<dim+natural; j++) {
            if (j == i) continue;
            x[j]++;
            f = func(dim, x);  oracle_call++;
            if (f < min_f) {
                min_f = f;  *plus = j;  *minus = i;
                improved = TRUE;  improve_call++;
            }
            x[j]--;
        }
        x[i]++;
    }
    return improved;
}


/* returns the minimum value of func(x)    */
/* INPUT  func   : pointer to the function */
/* INPUT  init[] : the initial solution    */
/* OUTPUT init[] : the minimized solution  */
static double minimize(int dim, double func(int,int*), int init[],
                       int natural) {
    int *x, plus, minus;

    counter_reset();
    if (!is_feasible(dim, func, init)) return PLUS_INFINITY;
    status_print(dim, func, init);

    assert(natural == 0 || natural == 1);
    x = ivector_new(dim + natural);
    ivector_copy(dim, x, init);

    while (best_improve(dim, func, x, &plus, &minus, natural)) {
        x[plus]++;
        x[minus]--;
        status_print(dim, func, x);
    }
    ivector_copy(dim, init, x);
    free(x);

    counter_print();
    return func(dim, init);
}


double mconv_minimize(int dim, double func(int,int*), int init[]) {
    return minimize(dim, func, init, 0);
}

double mgconv_minimize(int dim, double func(int,int*), int init[]) {
    return minimize(dim, func, init, 1);
}
