/* L# convex function -> submodular set function converter */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"
#include "sfm8.h"
#include "wolfe.h"

int submodular_type = IFF;

static double   stored_func_val;
static double (*stored_func)(int,int*);
static int     *stored_x;
static int      stored_direction;

static double oracle(int *set, int m, int n) {
    int i;
    double f;

    if (m == 0) return 0;
    for (i=0; i<m; i++) stored_x[set[i]] += stored_direction;
    f = stored_func(n, stored_x);
    for (i=0; i<m; i++) stored_x[set[i]] -= stored_direction;
    oracle_call++;
#ifdef DEBUG2
    printf("%.14f (%g - %g) ", f-stored_func_val, f, stored_func_val);
    for (i=0; i<m; i++) printf("[%d]", set[i]);
    printf("\n");
#endif /* DEBUG2 */
    return f - stored_func_val;
}

double sf_minimize(int dim, double func(int,int*), int x[], int direction,
                   int set[]) {
    stored_func_val  = func(dim, x);
    stored_func      = func;
    stored_x         = x;
    stored_direction = direction;
    double f1, f2;

    switch (submodular_type) {
    case IFF:   set[0] = sfm_call(dim, oracle, set+1); break;
    case WOLFE: set[0] = wolfe_call(dim, oracle, set+1); break;
    case IFF_WOLFE:
        set[0] = sfm_call(dim, oracle, set+1);
        f1 = oracle(set+1, set[0], dim);
        set[0] = wolfe_call(dim, oracle, set+1);
        f2 = oracle(set+1, set[0], dim);
        if (fcmp(f1,f2) != 0) {
            printf("iff = %f, wolfe = %f\n", f1, f2);
            exit(1);
        }
        printf("ans = "); ivector_print(set[0], set+1); printf("\n");
        break;
    default:    assert(0!=0); break;
    }
    return oracle(set+1, set[0], dim);
}
