#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "wolfe.h"
#include "odicon.h"
/*#include "quicksort.c"*/

#define TRUE  1
#define FALSE 0
#define N 1000

#define EPS 1.0E-10

double x[N+1];  /**  the minimum-norm base **/
int jmax,jmin;  /** the sizes of the max and the min minimizers **/
int smax[N+1], smin[N+1]; /** the indices of the max and the min minimizers **/
int number_extreme_point = 0; /** the number of generated extreme base **/



int
mnp2sfm(int n, double oracle(int *p, int j, int n))
{
  double gamma, fplus, fminus;
  double eps;
  int i,j;

  /* \sum { x(v) | x(v) < 0} */
  gamma=0.0;
  for (i = 1; i <= n; i++){
    if (x[i] < 0){
      gamma += x[i];
    }
    /**    printf("x[%d] = %8.15f\n",i, x[i]);**/ 
  }
    
  /* Tolerance for computing x to obtain minimizers when f is integral */    
  eps = (double) 1.0 / n;

  /* x^+ = {u | x(u) \leq 0} */
  for(i = 0; i <= n; i++){
    smax[i] = 0;
  }

  for (i = 1, j = 1 , jmax = 0; i <= n ; i++){
    if (x[i] <= eps){
      smax[j] = i-1;
      jmax = j; j++;
    }
  }
  fplus = oracle(smax+1, jmax, n);
  smax[0] = jmax;

  /* x^- = {u | x(u) < 0} */
  for(i = 0; i <= n; i++){
    smin[i] = 0;
  }
  for (i = 1, j = 1, jmin = 0 ; i <= n ; i++){
    if (x[i] < -eps){
      smin[j] = i-1; 
      jmin = j; j++;
    }
  }
  fminus = oracle(smin+1, jmin, n);  
  smin[0] = jmin;

#ifdef DEBUG
  if (fabs(fminus-gamma) > 0.5 || fabs(fplus-gamma) > 0.5 ) {
  printf("|max_minimizer| = %d\n",jmax);
  printf("|min_minimizer| = %d\n",jmin);
  printf("++++++++++++++++++++++++++\n");
  printf("x^-(V) = %8.15f\n", gamma);
  printf("f(x^+) = %8.15f\n", fplus);
  printf("f(x^-) = %8.15f\n", fminus);
   
  printf("\nMaximum Minimizer\n");
  printf("{");
  for (i=1; i<=jmax; i++){
    printf("%d", smax[i]);
    if (i!=jmax) printf(",");
  }
  printf("}\n");

  printf("\nMinimum Minimizer\n");
  printf("{");
  for (i=1; i<=jmin; i++){
    printf("%d", smin[i]);
    if (i!=jmin) printf(",");
  }
  printf("}\n");

  fvector_print(n, x+1);
  printf("\neps = %f\n", eps);
  }
#endif /* DEBUG */

  return ( fabs(fminus-fplus) <= 0.5 );
  /*return ( fabs(fminus-gamma) <= 0.5 && fabs(fplus-gamma) <= 0.5 );*/
}


static int cmp(const void *p, const void *q) {
    const int *a = p, *b = q;
    double diff = x[*a] - x[*b];

    if (diff < -EPS) return -1;
    if (diff > +EPS) return +1;
    return (*a - *b);  /* to keep stability */
}


int
wolfe(int n, double sf(int *p, int j, int n))
{
  int i, ih, it, ic, ix, ixh, ii, j, jj, kx;
  int jk, k, kk, ipk, flag, bool;
  int ip[N+2], ib[N+2];
  int s[N+1];
  double xz[N+1];
  double u[N+1], ubar[N+1];
  double v[N+1], w[N+1];
  double pss[N+1];
  double sf_new, sf_previous;
  double theta, pp, xx, xpj, rr, usum, aid, bid, cid, rid, xxp, ss;
  static double r[N+1][N+1];
  static double ps[N+1][N+1];
  
  number_extreme_point = 0;

  for (j = 1; j <= N + 1; j++){
    ip[j] = j+1;
    ib[j] = 0;
    u[j] = w[j] = 0;
  }
  ip[N+1] = 0;
  
  kx = 0;
  xxp = 1.0E+10;
  ixh = 0;
  
  for (j = 1; j <= n; j++){
    x[j]  = 0.0;
  }

  for(j = 0; j <= n; j++){
      s[j] = j;
  }
  /**************   step 0 *********************/
  xx = 1.0;
  sf_previous = 0.0;
  for (j = 1; j <= n ; j++){
    sf_new = sf(s,j,n);
    ps[1][j] = sf_new - sf_previous;
    sf_previous = sf_new;
    xx += ps[1][j]*ps[1][j];
  }
  pss[1] = xx - 1.0;
  r[1][1] = sqrt(xx);
  ih = it = w[1] = 1;
  ic = 2;
  ip[ih] = 0;
  k = 1;
  
  /**************   step 1 *********************/
  
  /******************* (A) ********************/
  
  while(1){/* 1000 */
    for (j = 1; j <= n; j++){
      xz[j]  = x[j];
      x[j] = 0.0;
    }
    
    i = ih;
    while( i != 0){
      for(j = 1; j <= n; j++){
	x[j]  += ps[i][j]*w[i];
      }
      i = ip[i];
    }

    /******************* (B) ********************/
    
    /* sort x[1..n] using quicksort algorithm:  
       s[i] contains the index of i-th smallest value of x */
    for(j = 0; j <= n; j++){
      s[j] = j;
    }   
    /*quicksort2(x,s,1,n);*/
    qsort(s+1, n, sizeof(*s), cmp);
    for (j = 1; j <= n; j++) s[j]--;  /* change to 0-origin */
    
    /* generate a new extreme base */
    sf_previous = 0.0;
    for (j = 1; j <= n ; j++){
      sf_new = sf(s+1,j,n);      
      ps[ic][s[j]+1] =  sf_new - sf_previous;
      sf_previous = sf_new;
    }
    
    number_extreme_point++;

    /******************* (C) ********************/
    xx = xpj = pp = 0.0;
    
    for (j = 1; j <= n; j++){
      pp += ps[ic][j] * ps[ic][j];
      xx += x[j]*x[j];
      xpj += x[j] * ps[ic][j];
    }
    pss[ic] = pp;
    
    if ( (xxp - xpj) <= EPS){
      kx++;
      if (kx >= 10) { printf("exit kx\n"); return 0; }
    }


    i = ih;
    ss = 1.0;
    while( i != 0){
      if (ss < pss[i]){ ss = pss[i];}
      i = ip[i];
    }
    if (ss < pp) ss = pp;


    xxp = xx;
    if ((xx - xpj) < (ss * EPS) ){ /*printf("exit xx = xpj, k=%d\n", k);*/ return 0; }

    if (k == n){
      printf("Trying to augment full-dimensional S!\n ");
      return (0);
    }
    
    /******************* (D) ********************/
  
    i = ih;
    while ( i != 0){
      flag = 0;
      for (j =  1; j <= n; j++){
	if( ((ps[i][j] - ps[ic][j])*(ps[i][j] - ps[ic][j])) > EPS ) {
	  i = ip[i];
	  flag = 1;
	  break;
	}
      }      
      if (!flag) { printf("The generated point is in S!\n"); return 0; }
    }
    
    /******************* (E), (F), (G) ********************/
    rr = 0.0;
    i  = ih;
    for ( jk = 1; jk <= k; jk++){
      r[ic][jk] = 1.0;
      for (j = 1; j  <= n; j++){
	r[ic][jk] += ps[i][j]*ps[ic][j];
      }
      for (jj = 1; jj <= jk -1; jj++){
	r[ic][jk] -=  r[i][jj]*r[ic][jj];
      }
      r[ic][jk]  = r[ic][jk]/r[i][jk];
      rr += r[ic][jk]*r[ic][jk];
      i = ip[i];
    }

    r[ic][k+1] = sqrt(1+pp-rr);
    
    k++;
    ip[it] = ic;
    ib[ic] = it;
    it = ic;
    ic = ip[ic];
    ip[it] = 0;

    w[it] = 0.0;

    
    
    /**************   step 2 *********************/
    while(1){ /* 2000 */
      i = ih;
      for (j = 1; j  <= k; j++){
	ubar[j] = 1.0;
	for (jj = 1;jj <= j-1; jj++){
	  ubar[j]  -=  ubar[jj]*r[i][jj];
	}
	ubar[j] = ubar[j]/r[i][j];
	i = ip[i];
      }
      
      usum = 0.0;
      i = it;
      for (j = k; j >= 1; j--){
	u[i] = ubar[j];
	ii = ip[i];
	while (ii != 0){
	  u[i] = u[i] - r[ii][j]*u[ii];
	  ii = ip[ii];
	}
	u[i] = u[i]/r[i][j];
	usum += u[i];
	i  = ib[i];
      }
      
      bool = TRUE;
      i = ih;
      while  (i != 0){
	v[i] = u[i]/usum;
	if (v[i] < EPS) bool = FALSE;
	i = ip[i];
      }
      
      if (bool) {
	i = ih;
	while (i != 0){
	  w[i] = v[i];
	  i = ip[i];
	}
	break;
      }
      /**************   step 3 *********************/
  
      /******************* (A), (B) ********************/
      theta = 1.0;
      i = ih;
      while (i != 0){
	if ( (w[i] -v[i]) > EPS)
	  theta = (theta <= w[i]/(w[i]- v[i]) )? theta : w[i]/(w[i] -v[i]) ;
	i = ip[i];
      }
      
      /******************* (C) ********************/
      kk = 0;
      i = ih;
      for (j=1; j <= k; j++){
	w[i] = (1-theta)*w[i] + theta*v[i];
	ipk = ip[i];
	
	/******************* (D), (E), (F), (G) ********************/
	
	if( w[i] < EPS ){
	  w[i] = 0.0;
	  kk++;
	  ii = ip[i];
	  for  (jj  = j -kk +1; jj <= k -kk; jj++){
	    aid = r[ii][jj];
	    bid  = r[ii][jj +1];
	    cid = sqrt (aid*aid + bid*bid);
	    ix = ii;
	    while (ix != 0){
	      rid = r[ix][jj];
	      r[ix][jj]  =  (aid * rid + bid * r[ix][jj+1])/ cid;
	      r[ix][jj+1]  =  (-bid * rid + aid * r[ix][jj+1])/ cid;
	      ix  = ip[ix];
	    }
	    ii = ip[ii];
	  }
	  
	  if ( i == ih){
	    ih = ip[i];
	    ib[ih] = 0;
	  }
	  else if (i == it){
	    it = ib[i];
	    ip[it] =  0;
	  }
	  else{
	    ip[ib[i]] = ip[i];
	    ib[ip[i]] = ib[i];
	  }
	  ip[i] = ic;
	  ib[i] = 0;
	  ic = i;
	}
	i = ipk;
      }
      k = k - kk;

    } /* goto 2000 */ 
    
  }  /* goto 1000 */   
  printf("exit goto 1000\n");
  return 0;
}


int wolfe_call(int n, double oracle(int *p, int j, int n), int *set) { 
    int succeeded;
    double fmin, fmax;

    wolfe(n, oracle);
    succeeded = mnp2sfm(n, oracle);
#ifdef DEBUG
    assert( succeeded );
#endif /* DEBUG */
    fmin = oracle(smin+1, smin[0], n);
    fmax = oracle(smax+1, smax[0], n);
    if (fmin < fmax) {
        memcpy(set, smin+1, sizeof(*set) * smin[0]);
        return smin[0];
    } else {
        memcpy(set, smax+1, sizeof(*set) * smax[0]);
        return smax[0];
    }
}

#ifdef TEST

double sf(int *pt, int m, int n)
{
  double c,f;
  int j;

  c = 0.0;
  for (j = 0; j < m; j++){ 
    c+= 5*pt[j] - 3*n;
  }

  f = (n - m + 5)*m - c; 

  return (f);
}

/* main */
int
main ()
{
  int n;

  cpu_time();	/* dummy */
  n  = N;

  wolfe(n,oracle);
  printf("****************\n");
  printf("N = %d\n",n);
  printf("time: %.2f (sec)\n", cpu_time() );
  printf("The number of generated extreme points = %d\n", number_extreme_point);  

  mnp2sfm(n,oracle);
  return 0;
}

#endif
